'''
Copyright (c) 2012-2013, Agora Games, LLC All rights reserved.

https://github.com/agoragames/kairos/blob/master/LICENSE.txt
'''
__version__ = "0.6.2"

from timeseries import Timeseries
from exceptions import *
